/****************************************************************************
    
    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#include "NewtonatorInstr2.h"
#include "VariSource.h"
#include <stk/Skini.h>

using namespace stk;
using namespace std;

const int NewtonatorInstr2::NUM_NOTES = 6;
const size_t NewtonatorInstr2::POLYPHONY = 5;

NewtonatorInstr2::NewtonatorInstr2(unsigned int channels)
: MAX_VAL(1.0), MIN_VAL(-1.0)
{
    for (int note = 0; note < NUM_NOTES; note++)
    {
        Newtonator2* newNewt = new Newtonator2(NULL, channels);
        VariSource* floorSource = NULL;
        VariSource* gravModSource = NULL;
        for (unsigned int chan = 0; chan < channels; chan++)
        {
            //Set defaults for wave generators for the floor and gravity modifier
            floorSource = newNewt->getFloorSources() + chan;
            floorSource->setFrequency(Midi2Pitch[chan + 20]);
            floorSource->setType(VariSource::SINE);
            floorSource->setAmplitude(MAX_VAL * .05);

            gravModSource = newNewt->getGravModSources() + chan;
            gravModSource->setFrequency(0);
            gravModSource->setAmplitude(0);
            gravModSource->setType(VariSource::SINE);
        }
        _newtonators.push_back(newNewt);
    }
    _currNewt = NULL;
    _currNewtIt = _newtonators.begin();
    addSampleRateAlert(this);
    initChannels(channels);
}

NewtonatorInstr2::~NewtonatorInstr2()
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        delete *(it++);
    }
}

void NewtonatorInstr2::initChannels(unsigned int numChannels)
{
    //Make sure the number is reflected in lastFrame_
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->initChannels(numChannels);
    }

    lastFrame_.resize(1, numChannels, 0.0);
}

void NewtonatorInstr2::sampleRateChanged(StkFloat newRate, StkFloat oldRate)
{
    //May need to do something here, I dunno yet
}

void NewtonatorInstr2::setAllAmpEnvTimes(StkFloat aTime, StkFloat dTime, StkFloat sLevel, StkFloat rTime)
{
    NewtCollection::iterator it = _newtonators.begin();
    ADSR* ampEnvs = NULL;
    unsigned int numChannels = channelsOut();

    while (it != _newtonators.end())
    {
        ampEnvs = (*(it++))->getAmpEnvs();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            ampEnvs[chan].setAllTimes(aTime, dTime, sLevel, rTime);
        }
    }
}

void NewtonatorInstr2::setAllGModEnvTimes(StkFloat aTime, StkFloat dTime, StkFloat sLevel, StkFloat rTime)
{
    NewtCollection::iterator it = _newtonators.begin();
    VariSource* gmodSources = NULL;
    unsigned int numChannels = channelsOut();

    while (it != _newtonators.end())
    {
        gmodSources = (*(it++))->getGravModSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            gmodSources[chan].getEnv().setAllTimes(aTime, dTime, sLevel, rTime);
        }
    }
}

StkFloat NewtonatorInstr2::tick(unsigned int channel)
{
    StkFloat ret = 0.0;
    PolyphonyQueue::iterator it = _activeNewts.begin();

    while (it != _activeNewts.end())
    {
        if ((*it)->isPlaying())
        {
            ret += (*it)->newtTick(channel);
            it++;
        }
        else
        {
            it = _activeNewts.erase(it);
        }
    }

    return ret;
}

void NewtonatorInstr2::noteOn(StkFloat frequency, StkFloat amplitude)
{
    if (amplitude == 0.0)
    {
        //Actually a note-off event
        noteOff(frequency, .5);
    }
    else
    {
        Newtonator2* newt = getNextAvailNewt();

        newt->setFrequency(frequency);
        newt->keyOn(amplitude);
        addActiveNewt(newt);
    }
}

void NewtonatorInstr2::noteOff(StkFloat amplitude)
{
    //Uh-oh, no idea which note is being turned off. Turn off the most
    //recent "on" note
    PolyphonyQueue::reverse_iterator it(_activeNewts.end());
    PolyphonyQueue::reverse_iterator itLast(_activeNewts.begin());

    if (it != itLast)
    {
        (*it)->keyOff(amplitude);
    }
}

void NewtonatorInstr2::noteOff(StkFloat frequency, StkFloat amplitude)
{
    //Turn off the most recent "on" note that matches the
    //specified frequency
    PolyphonyQueue::reverse_iterator it(_activeNewts.end());
    PolyphonyQueue::reverse_iterator itLast(_activeNewts.begin());

    while (it != itLast)
    {
        if ((*it)->getFrequency() == frequency)
        {
            (*it)->keyOff(amplitude);
            break;
        }
        it++;
    }
}

Newtonator2* NewtonatorInstr2::getNextAvailNewt()
{
    if (_currNewtIt == _newtonators.end())
        _currNewtIt = _newtonators.begin();

    return *(_currNewtIt++);
}

void NewtonatorInstr2::addActiveNewt(Newtonator2* newt)
{
    _currNewt = newt;
    _activeNewts.push_back(newt);
    if (_activeNewts.size() > POLYPHONY)
    {
        _activeNewts.pop_front();
    }
}

void NewtonatorInstr2::setFrequency(StkFloat frequency)
{
    if (_currNewt != NULL)
    {
        _currNewt->setFrequency(frequency);
    }
}

void NewtonatorInstr2::controlChange(int number, StkFloat value)
{
}

void NewtonatorInstr2::setGravityScale(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setGravityScale(val);
    }
}

void NewtonatorInstr2::setGravModFreq(stk::StkFloat freq)
{
    NewtCollection::iterator it = _newtonators.begin();
    unsigned int numChannels = channelsOut();
    VariSource* gravMods = NULL;

    while (it != _newtonators.end())
    {
        gravMods = (*(it++))->getGravModSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            gravMods[chan].setFrequency(freq);
        }
    }
}

void NewtonatorInstr2::setFloorScale(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setFloorScale(val);
    }
}

void NewtonatorInstr2::setClip(bool val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setClip(val);
    }
}

void NewtonatorInstr2::setGravModDepth(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setGravModDepth(val);
    }
}

void NewtonatorInstr2::setFloorType(VariSource::Type type)
{
    NewtCollection::iterator it = _newtonators.begin();
    unsigned int numChannels = channelsOut();
    VariSource* floorSources = NULL;

    while (it != _newtonators.end())
    {
        floorSources = (*(it++))->getFloorSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            floorSources[chan].setType(type);
        }
    }
}

void NewtonatorInstr2::setGravModType(VariSource::Type type)
{
    NewtCollection::iterator it = _newtonators.begin();
    unsigned int numChannels = channelsOut();
    VariSource* gravModSources = NULL;

    while (it != _newtonators.end())
    {
        gravModSources = (*(it++))->getGravModSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            gravModSources[chan].setType(type);
        }
    }
}

void NewtonatorInstr2::setFloorFreq(stk::StkFloat freq)
{
    NewtCollection::iterator it = _newtonators.begin();
    unsigned int numChannels = channelsOut();
    VariSource* floorSources = NULL;

    while (it != _newtonators.end())
    {
        floorSources = (*(it++))->getFloorSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            floorSources[chan].setFrequency(freq);
        }
    }
}
